from .support_functions import *
from .config import *
import sys
from .base_classes import *

GDSClassesMapping = {
    'code': Coding,
    'chain': string,
    'issued': instant,
    'actionResulting': ResourceReference,
    'concern': ResourceReference,
    'focus': code,
    'prefix': string,
    'valueAttachment': Attachment,
    'assessment': ResourceReference,
    'query': base64Binary,
    'issuer': ResourceReference,
    'valueIdentifier': Identifier,
    'compose': ValueSet_Compose,
    'exampleUri': uri,
    'equivalence': ConceptMapEquivalence,
    'referenceResource': ResourceReference,
    'DigestMethod': DigestMethodType,
    'bornPeriod': Period,
    'KeyValue': KeyValueType,
    'exampleRatio': Ratio,
    'ingredient': Substance_Ingredient,
    'updated': DateTimeType,
    'lowerLimit': decimal,
    'dispense': Supply_Dispense,
    'sopclass': oid,
    'reasonResource': ResourceReference,
    'choice': Coding,
    'Signature': SignatureType,
    'refusedIndicator': boolean,
    'maxLength': integer,
    'dischargeDiagnosis': ResourceReference,
    'encounter': ResourceReference,
    'reAdmission': boolean,
    'exampleSchedule': Schedule,
    'gender': CodeableConcept,
    'notes': string,
    'additionalInstructions': CodeableConcept,
    'item': ResourceReference,
    'valueRatio': Ratio,
    'Transforms': TransformsType,
    'answerString': string,
    'XPath': string,
    'requirements': string,
    'numberOfInstances': integer,
    'dataInstant': instant,
    'rate': Ratio,
    'asNeededCodeableConcept': CodeableConcept,
    'dataAttachment': Attachment,
    'section': Composition_Section,
    'X509Certificate': base64Binary,
    'version': string,
    'emptyReason': CodeableConcept,
    'nameReference': string,
    'supplier': ResourceReference,
    'method': CodeableConcept,
    'valueString': string,
    'hash': string,
    'relatesTo': DocumentReference_RelatesTo,
    'doseQuantity': Quantity,
    'reported': boolean,
    'dataHumanName': HumanName,
    'address': Address,
    'active': boolean,
    'presentedForm': Attachment,
    'documentMailbox': uri,
    'dataId': id,
    'publisher': string,
    'substance': ResourceReference,
    'remarks': string,
    'amount': Quantity,
    'published': DateTimeType,
    'action': SecurityEventAction,
    'SPKISexp': base64Binary,
    'options': ResourceReference,
    'answerDate': date,
    'dateAsserted': date,
    'multipleBirthInteger': integer,
    'diagnosticDateTime': dateTime,
    'confidentiality': CodeableConcept,
    'family': string,
    'PGPData': PGPDataType,
    'clinicalNotes': string,
    'codedDiagnosis': CodeableConcept,
    'exampleDate': date,
    'dataOid': oid,
    'exampleString': string,
    'total': integer,
    'use': NameUse,
    'reasonNotGiven': CodeableConcept,
    'zip': string,
    'destination': ResourceReference,
    'outcomeDesc': string,
    'maxDosePerPeriod': Ratio,
    'next': Extension,
    'exampleOid': oid,
    'answerDecimal': decimal,
    'extensionDefn': Profile_ExtensionDefn,
    'stage': Condition_Stage,
    'type': CodeableConcept,
    'dataUri': uri,
    'valueUri': uri,
    'relatedItem': Procedure_RelatedItem,
    'flag': CodeableConcept,
    'lotNumber': string,
    'exampleDateTime': dateTime,
    'valueDateTime': dateTime,
    'op': FilterOperator,
    'endpoint': uri,
    'policyManager': uri,
    'constraint': Profile_Constraint,
    'modified': dateTime,
    'longitude': decimal,
    'dosage': MedicationStatement_Dosage,
    'performer': Procedure_Performer,
    'abatementAge': Age,
    'orderedItem': ResourceReference,
    'agent': Provenance_Agent,
    'high': Quantity,
    'timingDateTime': dateTime,
    'expirationDate': date,
    'fulfillment': ResourceReference,
    'whenGiven': Period,
    'exclude': ValueSet_Include,
    'species': CodeableConcept,
    'suppliedItem': ResourceReference,
    'concept': ValueSet_Concept,
    'modifierExtension': Extension,
    'end': dateTime,
    'goal': CarePlan_Goal,
    'dataInteger': integer,
    'attachment': ResourceReference,
    'doseTarget': CodeableConcept,
    'X509IssuerName': string,
    'parameter': Extension,
    'profile': uri,
    'map': string,
    'product': Medication_Product,
    'description': string,
    'dataIdentifier': Identifier,
    'max': string,
    'mapping': Profile_Mapping1,
    'collection': Specimen_Collection,
    'accessionIdentifier': Identifier,
    'date': dateTime,
    'participant': SecurityEvent_Participant,
    'data': ResourceReference,
    'response': Query_Response,
    'X509CRL': base64Binary,
    'organization': ResourceReference,
    'object': SecurityEvent_Object,
    'documentation': string,
    'element': Profile_Element,
    'representation': PropertyRepresentation,
    'valueUuid': uuid,
    'order': ResourceReference,
    'origin': ResourceReference,
    'interpretation': CodeableConcept,
    'dataCoding': Coding,
    'telecom': Contact,
    'dataPeriod': Period,
    'userId': string,
    'entity': Provenance_Entity,
    'releaseDate': dateTime,
    'report': ResourceReference,
    'appliesPeriod': Period,
    'valueAddress': Address,
    'valueCodeableConcept': CodeableConcept,
    'group': Questionnaire_Group,
    'searchInclude': string,
    'KeyName': string,
    'valueBoolean': boolean,
    'latitude': decimal,
    'conformance': BindingConformance,
    'dataSchedule': Schedule,
    'animal': Patient_Animal,
    'specialCourtesy': CodeableConcept,
    'timestamp': instant,
    'primary': boolean,
    'valueRange': Range,
    'RSAKeyValue': RSAKeyValueType,
    'requestDetail': ResourceReference,
    'accomodation': Encounter_Accomodation,
    'valueId': id,
    'name': string,
    'clinicalInformation': string,
    'referrer': ResourceReference,
    'exampleResource': ResourceReference,
    'mode': LocationMode,
    'dependsOn': ConceptMap_DependsOn,
    'identifier': Identifier,
    'upperLimit': decimal,
    'series': string,
    'deceasedDate': date,
    'frames': integer,
    'frequency': integer,
    'exampleContact': Contact,
    'operation': Conformance_Operation,
    'collector': ResourceReference,
    'event': SecurityEvent_Event,
    'category': CategoryType,
    'managingOrganization': ResourceReference,
    'container': Specimen_Container,
    'network': SecurityEvent_Network,
    'dataCodeableConcept': CodeableConcept,
    'publish': boolean,
    'content': Medication_Content,
    'dataSampledData': SampledData,
    'contained': Resource_Inline,
    'issue': OperationOutcome_Issue,
    'model': string,
    'KeyInfo': KeyInfoType,
    'valueContact': Contact,
    'reason': CodeableConcept,
    'seriesDoses': integer,
    'definition': string,
    'language': code,
    'created': date,
    'route': CodeableConcept,
    'filter': ValueSet_Filter,
    'length': integer,
    'qualification': Practitioner_Qualification,
    'first': Extension,
    'onsetDate': date,
    'MgmtData': string,
    'suffix': string,
    'X509Data': X509DataType,
    'isExtensible': boolean,
    'coding': Coding,
    'number': integer,
    'RetrievalMethod': RetrievalMethodType,
    'bodySite': CodeableConcept,
    'owner': ResourceReference,
    'discriminator': id,
    'size': integer,
    'city': string,
    'given': string,
    'service': DocumentReference_Service,
    'width': integer,
    'breed': CodeableConcept,
    'system': uri,
    'lifecycle': SecurityEventObjectLifecycle,
    'priority': CodeableConcept,
    'isModifier': boolean,
    'sensitivity': CodeableConcept,
    'relationship': HierarchicalRelationshipType,
    'masterIdentifier': Identifier,
    'explanation': Immunization_Explanation,
    'answerInstant': instant,
    'short': string,
    'authorityResource': ResourceReference,
    'udi': string,
    'symptom': AdverseReaction_Symptom,
    'exampleCoding': Coding,
    'kind': CodeableConcept,
    'target': ResourceReference,
    'bed': ResourceReference,
    'dispenser': ResourceReference,
    'outcome': SecurityEventOutcome,
    'valueBase64Binary': base64Binary,
    'mimeType': code,
    'dataBase64Binary': base64Binary,
    'timingPeriod': Period,
    'caseSensitive': boolean,
    'exampleRange': Range,
    'timingSchedule': Schedule,
    'modality': Modality,
    'dailyAmount': Quantity,
    'dataRange': Range,
    'min': integer,
    'responsible': ResourceReference,
    'note': string,
    'instance': Substance_Instance,
    'answerDateTime': dateTime,
    'indexed': instant,
    'facilityType': CodeableConcept,
    'channel': DeviceObservationReport_Channel,
    'validityPeriod': Period,
    'fhirVersion': id,
    'dataCode': code,
    'denominator': Quantity,
    'Manifest': ManifestType,
    'vaccinationProtocol': Immunization_VaccinationProtocol,
    'position': Location_Position,
    'device': ResourceReference,
    'class': EncounterClass,
    'answerInteger': integer,
    'collectedDateTime': dateTime,
    'synonym': string,
    'observation': ResourceReference,
    'url': uri,
    'attester': Composition_Attester,
    'request': ResourceReference,
    'uri': UriType,
    'quantity': Quantity,
    'exampleQuantity': Quantity,
    'refusalReason': CodeableConcept,
    'text': string,
    'supercedes': ResourceReference,
    'isBrand': boolean,
    'X509IssuerSerial': X509IssuerSerialType,
    'relation': FamilyHistory_Relation,
    'dataDate': date,
    'availability': InstanceAvailability,
    'accessionNo': Identifier,
    'admitSource': CodeableConcept,
    'deceasedDateTime': dateTime,
    'copyright': string,
    'title': TextType,
    'implementation': Conformance_Implementation,
    'valueCode': code,
    'X509SerialNumber': integer,
    'factor': decimal,
    'diagnosticPeriod': Period,
    'doseStatus': CodeableConcept,
    'contentType': code,
    'custodian': ResourceReference,
    'exampleBoolean': boolean,
    'recordedDate': dateTime,
    'timing': Schedule,
    'wasNotGiven': boolean,
    'resource': Conformance_Resource,
    'dataString': string,
    'dateCriterion': ImmunizationRecommendation_DateCriterion,
    'summary': TextType,
    'subtype': CodeableConcept,
    'SignedInfo': SignedInfoType,
    'activity': CarePlan_Activity,
    'view': CodeableConcept,
    'reference': ResourceReference,
    'exampleAttachment': Attachment,
    'SignatureMethod': SignatureMethodType,
    'individual': ResourceReference,
    'result': ResourceReference,
    'interpreter': ResourceReference,
    'updateCreate': boolean,
    'primaryLanguage': code,
    'subject': ResourceReference,
    'followUp': string,
    'exampleSampledData': SampledData,
    'capacity': Quantity,
    'expectedSupplyDuration': Duration,
    'label': string,
    'state': string,
    'dataAddress': Address,
    'multipleBirthBoolean': boolean,
    'import': uri,
    'exampleInstant': instant,
    'PGPKeyPacket': base64Binary,
    'valueHumanName': HumanName,
    'exampleDecimal': decimal,
    'valueDate': date,
    'SignatureProperty': SignaturePropertyType,
    'valueSampledData': SampledData,
    'reaction': Immunization_Reaction,
    'whenPrepared': Period,
    'prescription': ResourceReference,
    'country': string,
    'responsibleParty': ResourceReference,
    'Object': ObjectType,
    'context': string,
    'receiver': ResourceReference,
    'experimental': boolean,
    'comment': string,
    'exampleUuid': uuid,
    'key': id,
    'simple': CarePlan_Simple,
    'dataContact': Contact,
    'valueDecimal': decimal,
    'period': Period,
    'birthDate': dateTime,
    'height': integer,
    'header': string,
    'onsetRange': Range,
    'path': string,
    'exampleAddress': Address,
    'onsetString': string,
    'abatementDate': date,
    'exposure': AdverseReaction_Exposure,
    'requestor': boolean,
    'enterer': ResourceReference,
    'didNotOccurFlag': boolean,
    'location': ResourceReference,
    'collectedPeriod': Period,
    'sensitivityTest': ResourceReference,
    'hospitalization': Encounter_Hospitalization,
    'policy': uri,
    'treatment': Specimen_Treatment,
    'last': Extension,
    'exampleCode': code,
    'valuePeriod': Period,
    'comparator': QuantityCompararator,
    'value': code,
    'deceasedBoolean': boolean,
    'contextType': ExtensionContext,
    'dischargeDisposition': CodeableConcept,
    'property': code,
    'procedure': CodeableConcept,
    'serviceCategory': CodeableConcept,
    'dataResource': ResourceReference,
    'causalityExpectation': CausalityExpectation,
    'metric': DeviceObservationReport_Metric,
    'binding': Profile_Binding,
    'site': string,
    'PGPKeyID': base64Binary,
    'numberOfSeries': integer,
    'valueResource': ResourceReference,
    'asNeededBoolean': boolean,
    'id': IdType,
    'feed': FeedType,
    'reasonCodeableConcept': CodeableConcept,
    'bornString': string,
    'severity': ConstraintSeverity,
    'author': PersonType,
    'media': Coding,
    'authorizingPrescription': ResourceReference,
    'member': ResourceReference,
    'rules': SlicingRules,
    'units': UnitsOfTime,
    'party': ResourceReference,
    'document': Conformance_Document,
    'status': ValueSetStatus,
    'practitioner': ResourceReference,
    'purpose': string,
    'identity': id,
    'slicing': Profile_Slicing,
    'CanonicalizationMethod': CanonicalizationMethodType,
    'X509SKI': base64Binary,
    'forecastStatus': CodeableConcept,
    'imagingStudy': ResourceReference,
    'dataDecimal': decimal,
    'person': ResourceReference,
    'contact': Patient_Contact,
    'extensible': boolean,
    'prescriber': ResourceReference,
    'doseSequence': integer,
    'entry': EntryType,
    'vaccineType': CodeableConcept,
    'maritalStatus': CodeableConcept,
    'protocol': ImmunizationRecommendation_Protocol,
    'assigner': ResourceReference,
    'genderStatus': CodeableConcept,
    'photo': Attachment,
    'dateWritten': dateTime,
    'rest': Conformance_Rest,
    'human': string,
    'onsetAge': Age,
    'previous': Extension,
    'bornDate': date,
    'exampleCodeableConcept': CodeableConcept,
    'deceasedString': string,
    'source': Specimen_Source,
    'blob': base64Binary,
    'valueCoding': Coding,
    'criticality': Criticality,
    'format': uri,
    'authorityCodeableConcept': CodeableConcept,
    'serviceProvider': ResourceReference,
    'specimen': ResourceReference,
    'aggregation': AggregationMode,
    'Transform': TransformType,
    'asserter': ResourceReference,
    'recipient': ResourceReference,
    'preAdmissionIdentifier': Identifier,
    'doseNumber': integer,
    'formal': string,
    'X509SubjectName': string,
    'referenceUri': uri,
    'substitution': MedicationPrescription_Substitution,
    'dateTime': instant,
    'referenceRange': Observation_ReferenceRange,
    'indication': CodeableConcept,
    'SignatureProperties': SignaturePropertiesType,
    'security': Conformance_Security,
    'deviceName': string,
    'deceasedAge': Age,
    'valueSet': ResourceReference,
    'authority': ResourceReference,
    'include': ValueSet_Include,
    'recommendation': ImmunizationRecommendation_Recommendation,
    'duration': decimal,
    'SignatureValue': SignatureValueType,
    'dataRatio': Ratio,
    'dimensions': integer,
    'altId': string,
    'recorded': instant,
    'comments': string,
    'recorder': ResourceReference,
    'answerBoolean': boolean,
    'valueSchedule': Schedule,
    'who': ResourceReference,
    'patient': ResourceReference,
    'schedule': Schedule,
    'deleted': boolean,
    'specialArrangement': CodeableConcept,
    'complication': CodeableConcept,
    'expansion': ValueSet_Expansion,
    'expiry': dateTime,
    'whenHandedOver': Period,
    'prohibited': boolean,
    'specimenQuantity': Quantity,
    'exampleInteger': integer,
    'cors': boolean,
    'additive': ResourceReference,
    'manufacturer': ResourceReference,
    'exampleBase64Binary': base64Binary,
    'actual': boolean,
    'certificate': Conformance_Certificate,
    'extension': Extension,
    'package': Medication_Package,
    'dataBoolean': boolean,
    'certainty': CodeableConcept,
    'numerator': Quantity,
    'numberOfRepeatsAllowed': integer,
    'valueQuantity': Quantity,
    'messaging': Conformance_Messaging,
    'Reference': ReferenceType,
    'software': string,
    'contains': ValueSet_Contains,
    'communication': CodeableConcept,
    'image': DiagnosticReport_Image,
    'readHistory': boolean,
    'exampleId': id,
    'operator': ResourceReference,
    'altitude': decimal,
    'question': Questionnaire_Question,
    'start': dateTime,
    'integritySignature': string,
    'low': Quantity,
    'acceptUnknown': boolean,
    'authored': dateTime,
    'xpath': string,
    'form': CodeableConcept,
    'specialty': CodeableConcept,
    'authenticator': ResourceReference,
    'link': linkType,
    'related': Observation_Related,
    'docStatus': CodeableConcept,
    'line': string,
    'DSAKeyValue': DSAKeyValueType,
    'count': integer,
    'careProvider': ResourceReference,
    'characteristic': Group_Characteristic,
    'valueInstant': instant,
    'dosageInstruction': MedicationPrescription_DosageInstruction,
    'reliableCache': integer,
    'searchParam': Profile_SearchParam,
    'mustSupport': boolean,
    'define': ValueSet_Define,
    'display': string,
    'partOf': ResourceReference,
    'dataUuid': uuid,
    'medication': ResourceReference,
    'ordered': boolean,
    'uid': oid,
    'abstract': boolean,
    'condition': id,
    'exampleIdentifier': Identifier,
    'evidence': Condition_Evidence,
    'supportingPatientInformation': ResourceReference,
    'deceasedRange': Range,
    'orderer': ResourceReference,
    'reliability': ObservationReliability,
    'SPKIData': SPKIDataType,
    'exampleHumanName': HumanName,
    'valueOid': oid,
    'virtualDevice': DeviceObservationReport_VirtualDevice,
    'when': Order_When,
    'detail': SecurityEvent_Detail,
    'actor': ResourceReference,
    'other': ResourceReference,
    'role': SecurityEventObjectRole,
    'details': string,
    'appliesDateTime': dateTime,
    'conclusion': string,
    'examplePeriod': Period,
    'repeat': Schedule_Repeat,
    'valueInteger': integer,
    'diet': CodeableConcept,
    'meaning': CodeableConcept,
    'dataQuantity': Quantity,
    'requester': ResourceReference,
    'abatementBoolean': boolean,
    'timingString': string,
    'structure': Profile_Structure,
    'supportingImmunization': ResourceReference,
    'physicalType': CodeableConcept,
    'sensitivityType': SensitivityType,
    'age': Range,
    'sourceSite': CodeableConcept,
    'doseStatusReason': CodeableConcept,
    'dataDateTime': dateTime,
    'time': dateTime,
    'receivedTime': dateTime,
}

def get_root_tag(node):
    tag = Tag_pattern_.match(node.tag).groups()[-1]
    rootClass = GDSClassesMapping.get(tag)
    if rootClass is None:
        rootClass = globals().get(tag)
    return tag, rootClass


def parse(inFileName, silence=False):
    doc = parsexml_(inFileName)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'Element'
        rootClass = Element
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('<?xml version="1.0" ?>\n')
        rootObj.export(
            sys.stdout, 0, name_=rootTag,
            namespacedef_='',
            pretty_print=True)
    return rootObj


def parseEtree(inFileName, silence=False):
    doc = parsexml_(inFileName)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'Element'
        rootClass = Element
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    mapping = {}
    rootElement = rootObj.to_etree(None, name_=rootTag, mapping_=mapping)
    reverse_mapping = rootObj.gds_reverse_node_mapping(mapping)
    if not silence:
        content = etree_.tostring(
            rootElement, pretty_print=True,
            xml_declaration=True, encoding="utf-8")
        sys.stdout.write(content)
        sys.stdout.write('\n')
    return rootObj, rootElement, mapping, reverse_mapping


def parseString(inString, silence=False):
    from StringIO import StringIO
    doc = parsexml_(StringIO(inString))
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'Element'
        rootClass = Element
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('<?xml version="1.0" ?>\n')
        rootObj.export(
            sys.stdout, 0, name_=rootTag,
            namespacedef_='')
    return rootObj


def parseLiteral(inFileName, silence=False):
    doc = parsexml_(inFileName)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'Element'
        rootClass = Element
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('#from fhirBase import *\n\n')
        sys.stdout.write('import fhirBase as model_\n\n')
        sys.stdout.write('rootObj = model_.rootClass(\n')
        rootObj.exportLiteral(sys.stdout, 0, name_=rootTag)
        sys.stdout.write(')\n')
    return rootObj

__all__ = [
    "Address",
    "AddressUse",
    "AdverseReaction",
    "AdverseReaction_Exposure",
    "AdverseReaction_Symptom",
    "Age",
    "AggregationMode",
    "Alert",
    "AlertStatus",
    "AllergyIntolerance",
    "Attachment",
    "BackboneElement",
    "Binary",
    "BindingConformance",
    "Block",
    "CanonicalizationMethodType",
    "CarePlan",
    "CarePlanActivityCategory",
    "CarePlanActivityStatus",
    "CarePlanGoalStatus",
    "CarePlanStatus",
    "CarePlan_Activity",
    "CarePlan_Goal",
    "CarePlan_Participant",
    "CarePlan_Simple",
    "CategoryType",
    "CausalityExpectation",
    "CodeableConcept",
    "Coding",
    "Composition",
    "CompositionAttestationMode",
    "CompositionStatus",
    "Composition_Attester",
    "Composition_Event",
    "Composition_Section",
    "ConceptMap",
    "ConceptMapEquivalence",
    "ConceptMap_Concept",
    "ConceptMap_DependsOn",
    "ConceptMap_Map",
    "Condition",
    "ConditionRelationshipType",
    "ConditionStatus",
    "Condition_Evidence",
    "Condition_Location",
    "Condition_RelatedItem",
    "Condition_Stage",
    "Conformance",
    "ConformanceEventMode",
    "ConformanceStatementStatus",
    "Conformance_Certificate",
    "Conformance_Document",
    "Conformance_Event",
    "Conformance_Implementation",
    "Conformance_Messaging",
    "Conformance_Operation",
    "Conformance_Operation1",
    "Conformance_Query",
    "Conformance_Resource",
    "Conformance_Rest",
    "Conformance_SearchParam",
    "Conformance_Security",
    "Conformance_Software",
    "ConstraintSeverity",
    "Contact",
    "ContactSystem",
    "ContactUse",
    "ContentType",
    "Count",
    "Criticality",
    "DSAKeyValueType",
    "DateTimeType",
    "Device",
    "DeviceObservationReport",
    "DeviceObservationReport_Channel",
    "DeviceObservationReport_Metric",
    "DeviceObservationReport_VirtualDevice",
    "DiagnosticOrder",
    "DiagnosticOrderPriority",
    "DiagnosticOrderStatus",
    "DiagnosticOrder_Event",
    "DiagnosticOrder_Item",
    "DiagnosticReport",
    "DiagnosticReportStatus",
    "DiagnosticReport_Image",
    "DigestMethodType",
    "Distance",
    "DocumentManifest",
    "DocumentMode",
    "DocumentReference",
    "DocumentReferenceStatus",
    "DocumentReference_Context",
    "DocumentReference_Parameter",
    "DocumentReference_RelatesTo",
    "DocumentReference_Service",
    "DocumentRelationshipType",
    "Duration",
    "Element",
    "Encounter",
    "EncounterClass",
    "EncounterState",
    "Encounter_Accomodation",
    "Encounter_Hospitalization",
    "Encounter_Location",
    "Encounter_Participant",
    "EntryType",
    "EventTiming",
    "ExposureType",
    "Extension",
    "ExtensionContext",
    "FamilyHistory",
    "FamilyHistory_Condition",
    "FamilyHistory_Relation",
    "FeedType",
    "FilterOperator",
    "Flow",
    "Group",
    "GroupType",
    "Group_Characteristic",
    "HierarchicalRelationshipType",
    "HumanName",
    "IdType",
    "Identifier",
    "IdentifierUse",
    "ImagingModality",
    "ImagingStudy",
    "ImagingStudy_Instance",
    "ImagingStudy_Series",
    "Immunization",
    "ImmunizationRecommendation",
    "ImmunizationRecommendation_DateCriterion",
    "ImmunizationRecommendation_Protocol",
    "ImmunizationRecommendation_Recommendation",
    "Immunization_Explanation",
    "Immunization_Reaction",
    "Immunization_VaccinationProtocol",
    "Inline",
    "InstanceAvailability",
    "IssueSeverity",
    "KeyInfoType",
    "KeyValueType",
    "LinkType",
    "List",
    "ListMode",
    "List_Entry",
    "Location",
    "LocationMode",
    "LocationStatus",
    "Location_Position",
    "ManifestType",
    "Media",
    "MediaType",
    "Medication",
    "MedicationAdministration",
    "MedicationAdministrationStatus",
    "MedicationAdministration_Dosage",
    "MedicationDispense",
    "MedicationDispenseStatus",
    "MedicationDispense_Dispense",
    "MedicationDispense_Dosage",
    "MedicationDispense_Substitution",
    "MedicationKind",
    "MedicationPrescription",
    "MedicationPrescriptionStatus",
    "MedicationPrescription_Dispense",
    "MedicationPrescription_DosageInstruction",
    "MedicationPrescription_Substitution",
    "MedicationStatement",
    "MedicationStatement_Dosage",
    "Medication_Content",
    "Medication_Ingredient",
    "Medication_Package",
    "Medication_Product",
    "MessageHeader",
    "MessageHeader_Destination",
    "MessageHeader_Response",
    "MessageHeader_Source",
    "MessageSignificanceCategory",
    "Modality",
    "Money",
    "NameUse",
    "Narrative",
    "NarrativeStatus",
    "ObjectType",
    "Observation",
    "ObservationRelationshipType",
    "ObservationReliability",
    "ObservationStatus",
    "Observation_ReferenceRange",
    "Observation_Related",
    "OperationOutcome",
    "OperationOutcome_Issue",
    "Order",
    "OrderOutcomeStatus",
    "OrderResponse",
    "Order_When",
    "Organization",
    "Organization_Contact",
    "Other",
    "PGPDataType",
    "Patient",
    "Patient_Animal",
    "Patient_Contact",
    "Patient_Link",
    "Period",
    "PersonType",
    "Practitioner",
    "Practitioner_Qualification",
    "Procedure",
    "ProcedureRelationshipType",
    "Procedure_Performer",
    "Procedure_RelatedItem",
    "Profile",
    "Profile_Binding",
    "Profile_Constraint",
    "Profile_Definition",
    "Profile_Element",
    "Profile_ExtensionDefn",
    "Profile_Mapping",
    "Profile_Mapping1",
    "Profile_Query",
    "Profile_SearchParam",
    "Profile_Slicing",
    "Profile_Structure",
    "Profile_Type",
    "PropertyRepresentation",
    "Provenance",
    "ProvenanceEntityRole",
    "Provenance_Agent",
    "Provenance_Entity",
    "Quantity",
    "QuantityCompararator",
    "Query",
    "QueryOutcome",
    "Query_Response",
    "Questionnaire",
    "QuestionnaireStatus",
    "Questionnaire_Group",
    "Questionnaire_Question",
    "RSAKeyValueType",
    "Range",
    "Ratio",
    "ReactionSeverity",
    "ReferenceType",
    "RelatedPerson",
    "Resource",
    "ResourceProfileStatus",
    "ResourceReference",
    "Resource_Inline",
    "ResponseType",
    "RestfulConformanceMode",
    "RestfulOperationSystem",
    "RestfulOperationType",
    "RetrievalMethodType",
    "SPKIDataType",
    "SampledData",
    "SampledDataDataType",
    "Schedule",
    "Schedule_Repeat",
    "SearchParamType",
    "SecurityEvent",
    "SecurityEventAction",
    "SecurityEventObjectLifecycle",
    "SecurityEventObjectRole",
    "SecurityEventObjectType",
    "SecurityEventOutcome",
    "SecurityEventParticipantNetworkType",
    "SecurityEvent_Detail",
    "SecurityEvent_Event",
    "SecurityEvent_Network",
    "SecurityEvent_Object",
    "SecurityEvent_Participant",
    "SecurityEvent_Source",
    "SensitivityStatus",
    "SensitivityType",
    "SignatureMethodType",
    "SignaturePropertiesType",
    "SignaturePropertyType",
    "SignatureType",
    "SignatureValueType",
    "SignedInfoType",
    "SlicingRules",
    "Specimen",
    "Specimen_Collection",
    "Specimen_Container",
    "Specimen_Source",
    "Specimen_Treatment",
    "Substance",
    "Substance_Ingredient",
    "Substance_Instance",
    "Supply",
    "SupplyDispenseStatus",
    "SupplyStatus",
    "Supply_Dispense",
    "TextType",
    "TransformType",
    "TransformsType",
    "UnitsOfTime",
    "UriType",
    "ValueSet",
    "ValueSetStatus",
    "ValueSet_Compose",
    "ValueSet_Concept",
    "ValueSet_Contains",
    "ValueSet_Define",
    "ValueSet_Expansion",
    "ValueSet_Filter",
    "ValueSet_Include",
    "X509DataType",
    "X509IssuerSerialType",
    "a",
    "a_content",
    "abbr",
    "acronym",
    "area",
    "b",
    "base64Binary",
    "bdo",
    "big",
    "blockquote",
    "boolean",
    "br",
    "caption",
    "cite",
    "code",
    "col",
    "colgroup",
    "date",
    "dateTime",
    "dd",
    "decimal",
    "deleted_entry",
    "dfn",
    "div",
    "dl",
    "dt",
    "em",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "hr",
    "i",
    "id",
    "img",
    "instant",
    "integer",
    "kbd",
    "li",
    "linkType",
    "map",
    "oid",
    "ol",
    "p",
    "pre",
    "pre_content",
    "q",
    "samp",
    "small",
    "span",
    "string",
    "strong",
    "sub",
    "sup",
    "table",
    "tbody",
    "td",
    "tfoot",
    "th",
    "thead",
    "tr",
    "tt",
    "ul",
    "uri",
    "uuid",
    "var"
]




import xml.dom.minidom
domimpl = xml.dom.minidom.getDOMImplementation()

import sys
import string
import urllib
import urllib.parse
from io import StringIO

from . import utils
from .constants import COLLECTION, OBJECT, DAV_PROPS, RT_ALLPROP, RT_PROPNAME, RT_PROP
from .errors import *
from .utils import create_treelist, quote_uri, gen_estring

class COPY:
    """ copy resources and eventually create multistatus responses

    This module implements the COPY class which is responsible for
    copying resources. Usually the normal copy work is done in the
    interface class. This class only creates error messages if error
    occur.

    """


    def __init__(self,dataclass,src_uri,dst_uri,overwrite):
        self.__dataclass=dataclass
        self.__src=src_uri
        self.__dst=dst_uri
        self.__overwrite=overwrite


    def single_action(self):
        """ copy a normal resources.

        We try to copy it and return the result code.
        This is for Depth==0

        """

        dc=self.__dataclass
        base=self.__src

        ### some basic tests
        # test if dest exists and overwrite is false
        if dc.exists(self.__dst) and not self.__overwrite: raise DAV_Error(412)
        # test if src and dst are the same
        # (we assume that both uris are on the same server!)
        ps=urllib.parse.urlparse(self.__src)[2]
        pd=urllib.parse.urlparse(self.__dst)[2]
        if ps==pd: raise DAV_Error(403)

        return dc.copyone(self.__src,self.__dst,self.__overwrite)

        #return copyone(dc,self.__src,self.__dst,self.__overwrite)

    def tree_action(self):
        """ copy a tree of resources (a collection)

        Here we return a multistatus xml element.

        """
        dc=self.__dataclass
        base=self.__src

        ### some basic tests
        # test if dest exists and overwrite is false
        if dc.exists(self.__dst) and not self.__overwrite: raise DAV_Error(412)
        # test if src and dst are the same
        # (we assume that both uris are on the same server!)
        ps=urllib.parse.urlparse(self.__src)[2]
        pd=urllib.parse.urlparse(self.__dst)[2]
        if ps==pd: raise DAV_Error(403)
        
        
        result=dc.copytree(self.__src,self.__dst,self.__overwrite)
        #result=copytree(dc,self.__src,self.__dst,self.__overwrite)

        if not result: return None

        ###
        ### create the multistatus XML element
        ### (this is also the same as in delete.py.
        ###  we might make a common method out of it)
        ###

        doc = domimpl.createDocument(None, "D:multistatus", None)
        doc.documentElement.setAttribute("xmlns:D","DAV:")

        for el,ec in list(result.items()):
                re=doc.createElement("D:response")
                hr=doc.createElement("D:href")
                st=doc.createElement("D:status")
                huri=doc.createTextNode(quote_uri(el))
                t=doc.createTextNode(gen_estring(ec))
                st.appendChild(t)
                hr.appendChild(huri)
                re.appendChild(hr)
                re.appendChild(st)
                ms.appendChild(re)

        return doc.toxml(encoding="utf-8")



#!/usr/bin/env python
# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from argparse import ArgumentParser
from urllib.parse import urlparse
import unittest
import sys
import datetime

from dateutil.relativedelta import relativedelta
import caldav
import vobject
from proteus import config, Model, Wizard


def activate_module(name, config):
    Module = Model.get('ir.module')
    module, = Module.find([('name', '=', name)])
    if module.state != 'activated':
        Module.install([module.id], config.context)
        Wizard('ir.module.activate_upgrade').execute('upgrade')


def configure_user(login, config):
    User = Model.get('res.user')
    users = User.find([('login', '=', login)])
    email = '%s@example.com' % login
    if not users:
        User(login=login, name=login, email=email).save()
    else:
        user, = users
        user.email = email
        user.save()


def create_calendar(login, user, config):
    Calendar = Model.get('calendar.calendar')
    User = Model.get('res.user')
    current_user, = User.find([('login', '=', login)])
    main_user, = User.find([('login', '=', user)])
    for calendar in Calendar.find([('owner', '=', login)]):
        calendar.write_users.append(User(main_user.id))
        calendar.save()
        calendar.delete()
    calendar = Calendar(name=login, owner=current_user)
    if login != 'bar':
        calendar.read_users.append(main_user)
    calendar.save()


class TestCase(unittest.TestCase):

    def setUp(self):
        self.client = caldav.DAVClient(URL)
        self.principal = caldav.Principal(self.client, URL)
        self.calendar, = [x for x in self.principal.calendars()
            if x.url.path.endswith(user)]

    def test0010calendar(self):
        'Test calendar'
        calendar, _ = self.principal.calendars()

    def test0020create_event(self):
        'Create event'
        ical = vobject.iCalendar()
        vevent = ical.add('vevent')
        vevent.add('summary')
        vevent.summary.value = 'Test event'
        vevent.add('dtstart')
        vevent.dtstart.value = (datetime.datetime.now()
            + relativedelta(months=1))
        vevent.add('dtend')
        vevent.dtend.value = datetime.datetime.now() + relativedelta(months=1,
            hours=1)
        caldav.Event(self.client, data=ical.serialize(),
            parent=self.calendar).save()

    def test0030search_event(self):
        'Search date'
        events = self.calendar.date_search(datetime.datetime.now(),
            datetime.datetime.now() + relativedelta(months=2))
        self.assertEqual(len(events), 1)
        events = self.calendar.date_search(
            datetime.datetime.now() - relativedelta(months=1),
            datetime.datetime.now())
        self.assertEqual(len(events), 0)

    def test0040get_event(self):
        'Get event'
        event = self.calendar.events()[0]
        event.load()
        self.assertEqual(event.url, self.calendar.event(event.id).url)

    def test0050create_event_attendee(self):
        'Create event with attendee'
        ical = vobject.iCalendar()
        vevent = ical.add('vevent')
        vevent.add('summary')
        vevent.summary.value = 'Test event with attendee'
        vevent.add('dtstart')
        vevent.dtstart.value = datetime.datetime.now() + relativedelta(days=10)
        vevent.add('dtend')
        vevent.dtend.value = datetime.datetime.now() + relativedelta(days=10,
            hours=4)
        vevent.add('organizer')
        vevent.organizer.value = '%s@example.com' % user
        attendees = []
        for name in ('foo', 'bar'):
            attendee = vobject.base.ContentLine('ATTENDEE', [], '')
            attendee.partstat_param = 'TENTATIVE'
            attendee.value = 'MAILTO:%s@example.com' % name
            attendees.append(attendee)
        vevent.attendee_list = attendees
        caldav.Event(self.client, data=ical.serialize(),
            parent=self.calendar).save()

        Event = Model.get('calendar.event')
        owner_event, = Event.find([
                ('calendar.owner.email', '=', '%s@example.com' % user),
                ('summary', '=', vevent.summary.value),
                ])
        attendee_event, = Event.find([
                ('calendar.owner.email', '=', 'foo@example.com'),
                ])
        self.assertEqual(attendee_event.uuid, owner_event.uuid)

    def test0060update_attendee_status(self):
        'Update status of attendee'
        for event in self.calendar.events():
            event.load()
            value = event.instance.vevent.summary.value
            if value == 'Test event with attendee':
                break
        for attendee in event.instance.vevent.attendee_list:
            attendee.partstat_param = 'accepted'
        event.save()

        Event = Model.get('calendar.event')
        attendee_event, = Event.find([
                ('calendar.owner.email', '=', 'foo@example.com'),
                ])
        for attendee in attendee_event.attendees:
            self.assertEqual(attendee.status, 'accepted')

    def test0070delete_attendee(self):
        'Delete attendee'
        for event in self.calendar.events():
            event.load()
            value = event.instance.vevent.summary.value
            if value == 'Test event with attendee':
                break
        event.instance.vevent.attendee_list = []
        event.save()

        Event = Model.get('calendar.event')
        self.assertEqual(Event.find([
                    ('calendar.owner.email', '=', 'foo@example.com'),
                    ]), [])

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('--xmlrpc', dest='xmlrpc', metavar='URL',
        help='use trytond XML-RPC at URL')
    parser.add_argument('--url', dest='url', metavar='URL',
        help='use calendar at URL')
    options = parser.parse_args()
    config = config.set_xmlrpc(options.xmlrpc)
    xmlrpc_user = urlparse(options.xmlrpc).username
    user = urlparse(options.url).username
    assert xmlrpc_user == user
    assert user != 'foo'
    activate_module('calendar', config)
    configure_user(user, config)
    configure_user('foo', config)
    configure_user('bar', config)
    create_calendar(user, user, config)
    create_calendar('foo', user, config)
    create_calendar('bar', user, config)
    URL = options.url
    unittest.main(argv=sys.argv[:1])



from .base_classes import *
from .support_functions import *


class Other(Resource):
    """Other is a conformant for handling resource concepts not yet defined
    for FHIR or outside HL7's scope of interest.If the element is
    present, it must have either a @value, an @id, or extensions"""
    subclass = None
    superclass = Resource
    def __init__(self, id=None, extension=None, modifierExtension=None, language=None, text=None, contained=None, identifier=None, code=None, subject=None, author=None, created=None):
        self.original_tagname_ = None
        super(Other, self).__init__(id, extension, modifierExtension, language, text, contained, )
        if identifier is None:
            self.identifier = []
        else:
            self.identifier = identifier
        self.code = code
        self.subject = subject
        self.author = author
        self.created = created
    def factory(*args_, **kwargs_):
        if Other.subclass:
            return Other.subclass(*args_, **kwargs_)
        else:
            return Other(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_identifier(self): return self.identifier
    def set_identifier(self, identifier): self.identifier = identifier
    def add_identifier(self, value): self.identifier.append(value)
    def insert_identifier(self, index, value): self.identifier[index] = value
    def get_code(self): return self.code
    def set_code(self, code): self.code = code
    def get_subject(self): return self.subject
    def set_subject(self, subject): self.subject = subject
    def get_author(self): return self.author
    def set_author(self, author): self.author = author
    def get_created(self): return self.created
    def set_created(self, created): self.created = created
    def hasContent_(self):
        if (
            self.identifier or
            self.code is not None or
            self.subject is not None or
            self.author is not None or
            self.created is not None or
            super(Other, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Other', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Other')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Other', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Other'):
        super(Other, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Other')
    def exportChildren(self, outfile, level, namespace_='', name_='Other', fromsubclass_=False, pretty_print=True):
        super(Other, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        for identifier_ in self.identifier:
            identifier_.export(outfile, level, namespace_, name_='identifier', pretty_print=pretty_print)
        if self.code is not None:
            self.code.export(outfile, level, namespace_, name_='code', pretty_print=pretty_print)
        if self.subject is not None:
            self.subject.export(outfile, level, namespace_, name_='subject', pretty_print=pretty_print)
        if self.author is not None:
            self.author.export(outfile, level, namespace_, name_='author', pretty_print=pretty_print)
        if self.created is not None:
            self.created.export(outfile, level, namespace_, name_='created', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Other', mapping_=None):
        element = super(Other, self).to_etree(parent_element, name_, mapping_)
        for identifier_ in self.identifier:
            identifier_.to_etree(element, name_='identifier', mapping_=mapping_)
        if self.code is not None:
            code_ = self.code
            code_.to_etree(element, name_='code', mapping_=mapping_)
        if self.subject is not None:
            subject_ = self.subject
            subject_.to_etree(element, name_='subject', mapping_=mapping_)
        if self.author is not None:
            author_ = self.author
            author_.to_etree(element, name_='author', mapping_=mapping_)
        if self.created is not None:
            created_ = self.created
            created_.to_etree(element, name_='created', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Other, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'identifier':
            obj_ = Identifier.factory()
            obj_.build(child_)
            self.identifier.append(obj_)
            obj_.original_tagname_ = 'identifier'
        elif nodeName_ == 'code':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.code = obj_
            obj_.original_tagname_ = 'code'
        elif nodeName_ == 'subject':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.subject = obj_
            obj_.original_tagname_ = 'subject'
        elif nodeName_ == 'author':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.author = obj_
            obj_.original_tagname_ = 'author'
        elif nodeName_ == 'created':
            obj_ = date.factory()
            obj_.build(child_)
            self.created = obj_
            obj_.original_tagname_ = 'created'
        super(Other, self).buildChildren(child_, node, nodeName_, True)
# end class Other



# -*- coding: utf-8 -*-
##############################################################################
#
#    GNU Health: The Free Health and Hospital Information System
#    Copyright (C) 2008-2017 Luis Falcon <lfalcon@gnusolidario.org>
#    Copyright (C) 2011-2017 GNU Solidario <health@gnusolidario.org>
#
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import datetime
import decimal
from trytond.model import ModelView
from trytond.wizard import Wizard, StateTransition, StateView, Button
from trytond.transaction import Transaction
from trytond.pool import Pool


__all__ = ['CreateServiceInvoice']


class CreateServiceInvoice(Wizard):
    __name__ = 'gnuhealth.service.invoice.create'

#  
#  name: CreateServiceInvoice.discount_policy
#  @param : Insurance, service line product
#  @return : Policy applied to that service line
#  
    def discount_policy(self, insurance, product):
        # Check that there is a plan associated to the insurance
        if insurance.plan_id:
            # Traverse the product policies within the plan
            # In terms of applying discount, category and product are
            # mutually exclusive.
            
            discount = {}
            if insurance.plan_id.product_policy:
                for policy in insurance.plan_id.product_policy:
                    # Check first for product
                    if (product == policy.product):
                        if policy.discount:
                            discount['value'] = policy.discount
                            discount['type'] = 'pct'
                            return discount
                    
                for policy in insurance.plan_id.product_policy:
                    # Then, if there's no product, check for the category
                    if (policy.product_category in product.categories):
                        if policy.discount:
                            discount['value'] = policy.discount
                            discount['type'] = 'pct' 
                            return discount

            return discount
    
    def transition_create_service_invoice(self):
        pool = Pool()
        HealthService = pool.get('gnuhealth.health_service')
        Invoice = pool.get('account.invoice')
        Party = pool.get('party.party')
        Journal = pool.get('account.journal')

        currency_id = Transaction().context.get('currency')

        services = HealthService.browse(Transaction().context.get(
            'active_ids'))
        invoices = []

        #Invoice Header
        for service in services:
            if service.state == 'invoiced':
                self.raise_user_error('duplicate_invoice')
            if service.invoice_to:
                party = service.invoice_to
            else:
                party = service.patient.name
            invoice_data = {}
            invoice_data['description'] = service.desc
            invoice_data['party'] = party.id
            invoice_data['type'] = 'out'
            invoice_data['invoice_date'] = datetime.date.today()
            invoice_data['account'] = party.account_receivable.id
            invoice_data['company'] = service.company.id

            ctx = {}
            sale_price_list = None
            if hasattr(party, 'sale_price_list'):
                sale_price_list = party.sale_price_list

            if sale_price_list:
                ctx['price_list'] = sale_price_list.id
                ctx['sale_date'] = datetime.date.today()
                ctx['currency'] = currency_id
                ctx['customer'] = party.id

            journals = Journal.search([
                ('type', '=', 'revenue'),
                ], limit=1)

            if journals:
                journal, = journals
            else:
                journal = None

            invoice_data['journal'] = journal.id

            party_address = Party.address_get(party, type='invoice')
            if not party_address:
                self.raise_user_error('no_invoice_address')
            invoice_data['invoice_address'] = party_address.id
            invoice_data['reference'] = service.name

            if not party.customer_payment_term:
                self.raise_user_error('no_payment_term')

            invoice_data['payment_term'] = party.customer_payment_term.id

            #Invoice Lines
            seq = 0
            invoice_lines = []
            for line in service.service_line:
                seq = seq + 1
                account = line.product.template.account_revenue_used.id

                if sale_price_list:
                    with Transaction().set_context(ctx):
                        unit_price = sale_price_list.compute(party,
                            line.product, line.product.list_price,
                            line.qty, line.product.default_uom)
                else:
                    unit_price = line.product.list_price
                    
                if line.to_invoice:
                    taxes = []
                    desc = line.desc
                    
                    #Include taxes related to the product on the invoice line
                    for product_tax_line in line.product.customer_taxes_used:
                        taxes.append(product_tax_line.id)

                    # Check the Insurance policy for this service
                    if service.insurance_plan:
                        discount = \
                            self.discount_policy(service.insurance_plan, \
                                line.product)

                        if discount:
                            if 'value' in list(discount.keys()):
                                if discount['value']:
                                    if (discount['type'] == 'pct'):
                                        unit_price *= (1 - 
                                            decimal.Decimal(discount['value'])/100)

                                        # Add a remark on the description discount
                                        str_disc = str(discount['value']) + '%'
                                        desc = line.desc + " (Discnt " + \
                                          str (str_disc) + ")"

                    invoice_lines.append(('create', [{
                            'origin': str(line),
                            'product': line.product.id,
                            'description': desc,
                            'quantity': line.qty,
                            'account': account,
                            'unit': line.product.default_uom.id,
                            'unit_price': unit_price,
                            'sequence': seq,
                            'taxes': [('add',taxes)],
                        }]))
                invoice_data['lines'] = invoice_lines

            invoices.append(invoice_data)

        Invoice.update_taxes(Invoice.create(invoices))

        # Change to invoiced the status on the service document.
        HealthService.write(services, {'state': 'invoiced'})

        return 'end'



from .base_classes import *
from .support_functions import *


class Narrative(Element):
    """A human-readable formatted text, including images.If the element is
    present, it must have a value for at least one of the defined
    elements, an @id referenced from the Narrative, or extensions"""
    subclass = None
    superclass = Element
    def __init__(self, id=None, extension=None, status=None, div=None):
        self.original_tagname_ = None
        super(Narrative, self).__init__(id, extension, )
        self.status = status
        self.div = div
    def factory(*args_, **kwargs_):
        if Narrative.subclass:
            return Narrative.subclass(*args_, **kwargs_)
        else:
            return Narrative(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_status(self): return self.status
    def set_status(self, status): self.status = status
    def get_div(self): return self.div
    def set_div(self, div): self.div = div
    def hasContent_(self):
        if (
            self.status is not None or
            self.div is not None or
            super(Narrative, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Narrative', namespacedef_=' xmlns:xhtml="http://www.w3.org/1999/xhtml" ', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Narrative')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Narrative', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Narrative'):
        super(Narrative, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Narrative')
    def exportChildren(self, outfile, level, namespace_='', name_='Narrative', fromsubclass_=False, pretty_print=True):
        super(Narrative, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.status is not None:
            self.status.export(outfile, level, namespace_, name_='status', pretty_print=pretty_print)
        if self.div is not None:
            self.div.export(outfile, level, namespace_='xhtml:', name_='div', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Narrative', mapping_=None):
        element = super(Narrative, self).to_etree(parent_element, name_, mapping_)
        if self.status is not None:
            status_ = self.status
            status_.to_etree(element, name_='status', mapping_=mapping_)
        if self.div is not None:
            div_ = self.div
            div_.to_etree(element, name_='div', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Narrative, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'status':
            obj_ = NarrativeStatus.factory()
            obj_.build(child_)
            self.status = obj_
            obj_.original_tagname_ = 'status'
        elif nodeName_ == 'div':
            obj_ = div.factory()
            obj_.build(child_)
            self.div = obj_
            obj_.original_tagname_ = 'div'
        super(Narrative, self).buildChildren(child_, node, nodeName_, True)
# end class Narrative


class NarrativeStatus(Element):
    """The status of a resource narrativeIf the element is present, it must
    have either a @value, an @id, or extensions"""
    subclass = None
    superclass = Element
    def __init__(self, id=None, extension=None, value=None):
        self.original_tagname_ = None
        super(NarrativeStatus, self).__init__(id, extension, )
        self.value = _cast(None, value)
    def factory(*args_, **kwargs_):
        if NarrativeStatus.subclass:
            return NarrativeStatus.subclass(*args_, **kwargs_)
        else:
            return NarrativeStatus(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_value(self): return self.value
    def set_value(self, value): self.value = value
    def hasContent_(self):
        if (
            super(NarrativeStatus, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='NarrativeStatus', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='NarrativeStatus')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='NarrativeStatus', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='NarrativeStatus'):
        super(NarrativeStatus, self).exportAttributes(outfile, level, already_processed, namespace_, name_='NarrativeStatus')
        if self.value is not None and 'value' not in already_processed:
            already_processed.add('value')
            outfile.write(' value=%s' % (quote_attrib(self.value), ))
    def exportChildren(self, outfile, level, namespace_='', name_='NarrativeStatus', fromsubclass_=False, pretty_print=True):
        super(NarrativeStatus, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='NarrativeStatus', mapping_=None):
        element = super(NarrativeStatus, self).to_etree(parent_element, name_, mapping_)
        if self.value is not None:
            element.set('value', self.value)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('value', node)
        if value is not None and 'value' not in already_processed:
            already_processed.add('value')
            self.value = value
        super(NarrativeStatus, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        super(NarrativeStatus, self).buildChildren(child_, node, nodeName_, True)
        pass
# end class NarrativeStatus




import time
import re

from xml.dom import minidom

import urllib.request, urllib.parse, urllib.error
import urllib.parse

from .constants import RT_ALLPROP, RT_PROPNAME, RT_PROP
from http.server import BaseHTTPRequestHandler

def gen_estring(ecode):
    """ generate an error string from the given code """
    ec=int(ecode)
    if ec in BaseHTTPRequestHandler.responses:
        return "HTTP/1.1 %s %s" %(ec, BaseHTTPRequestHandler.responses[ec][0])
    else:
        return "HTTP/1.1 %s" %(ec)

def parse_propfind(xml_doc):
    """ 
    Parse an propfind xml file and return a list of props 
    """

    doc = minidom.parseString(xml_doc)

    request_type=None
    props={}
    namespaces=[]

    if doc.getElementsByTagNameNS("DAV:", "allprop"):
        request_type = RT_ALLPROP
    elif doc.getElementsByTagNameNS("DAV:", "propname"):
        request_type = RT_PROPNAME
    else:
        request_type = RT_PROP
        for i in doc.getElementsByTagNameNS("DAV:", "prop"):
            for e in i.childNodes:
                if e.nodeType != minidom.Node.ELEMENT_NODE:
                    continue
                ns = e.namespaceURI
                ename = e.localName
                if ns in props:
                    props[ns].append(ename)
                else:
                    props[ns]=[ename]
                    namespaces.append(ns)

    return request_type,props,namespaces


def create_treelist(dataclass,uri):
    """ create a list of resources out of a tree 

    This function is used for the COPY, MOVE and DELETE methods

    uri - the root of the subtree to flatten

    It will return the flattened tree as list

    """
    queue=[uri]
    list=[uri]
    while len(queue):
        element=queue[-1]
        if dataclass.is_collection(element):
            childs=dataclass.get_childs(element)
        else:
            childs=[]
        if len(childs):
            list=list+childs
        # update queue
        del queue[-1]
        if len(childs):
            queue=queue+childs
    return list

def is_prefix(uri1,uri2):
    """ returns 1 of uri1 is a prefix of uri2 """
    if uri2[:len(uri1)]==uri1:
        return 1
    else:
        return None

def quote_uri(uri):
    """ quote an URL but not the protocol part """
    up=urllib.parse.urlparse(uri)
    # np=urllib.parse.quote(up[2]).encode('utf-8')
    np=urllib.parse.quote(up[2])
    return urllib.parse.urlunparse((up[0],up[1],np,up[3],up[4],up[5]))

def get_uriparentpath(uri):
    """ extract the uri path and remove the last element """
    up=urllib.parse.urlparse(uri)
    return "/".join(up[2].split("/")[:-1])

def get_urifilename(uri):
    """ extract the uri path and return the last element """
    up=urllib.parse.urlparse(uri)
    res = up[2]
    res = res.split("/")[-1]
    return res

def get_parenturi(uri):
    """ return the parent of the given resource"""
    up=urllib.parse.urlparse(uri).decode()

    np='/'.join(up[2].split("/")[:-1])
    return urllib.parse.urlunparse((up[0],up[1],np,up[3],up[4],up[5]))

### XML utilities

def make_xmlresponse(result):
    """ construct a response from a dict of uri:error_code elements """
    doc = minidom.getDOMImplementation().createDocument(None, "multistatus", None)
    doc.documentElement.setAttribute("xmlns:D","DAV:")
    doc.documentElement.tagName = "D:multistatus"

    for el,ec in list(result.items()):
        re=doc.createElementNS("DAV:","response")
        hr=doc.createElementNS("DAV:","href")
        st=doc.createElementNS("DAV:","status")
        huri=doc.createTextNode(quote_uri(el))
        t=doc.createTextNode(gen_estring(ec))
        st.appendChild(t)
        hr.appendChild(huri)
        re.appendChild(hr)
        re.appendChild(st)
        doc.documentElement.appendChild(re)

    return doc.toxml(encoding="utf-8")

# taken from App.Common

weekday_abbr = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
weekday_full = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
                'Friday', 'Saturday', 'Sunday']
monthname    = [None, 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

def rfc1123_date(ts=None):
    # Return an RFC 1123 format date string, required for
    # use in HTTP Date headers per the HTTP 1.1 spec.
    # 'Fri, 10 Nov 2000 16:21:09 GMT'
    if ts is None: ts=time.time()
    year, month, day, hh, mm, ss, wd, y, z = time.gmtime(ts)
    return "%s, %02d %3s %4d %02d:%02d:%02d GMT" % (weekday_abbr[wd],
                                                    day, monthname[month],
                                                    year,
                                                    hh, mm, ss)
def iso8601_date(ts=None):
    # Return an ISO 8601 formatted date string, required
    # for certain DAV properties.
    # '2000-11-10T16:21:09-08:00
    if ts is None: ts=time.time()
    return time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime(ts))

def rfc850_date(ts=None):
    # Return an HTTP-date formatted date string.
    # 'Friday, 10-Nov-00 16:21:09 GMT'
    if ts is None: ts=time.time()
    year, month, day, hh, mm, ss, wd, y, z = time.gmtime(ts)
    return "%s, %02d-%3s-%2s %02d:%02d:%02d GMT" % (
            weekday_full[wd],
            day, monthname[month],
            str(year)[2:],
            hh, mm, ss)

### If: header handling support.  IfParser returns a sequence of
### TagList objects in the order they were parsed which can then
### be used in WebDAV methods to decide whether an operation can
### proceed or to raise HTTP Error 412 (Precondition failed)
IfHdr = re.compile(
    r"(?P<resource><.+?>)?\s*\((?P<listitem>[^)]+)\)"
    )

ListItem = re.compile(
    r"(?P<not>not)?\s*(?P<listitem><[a-zA-Z]+:[^>]*>|\[.*?\])",
    re.I)

class TagList:
    def __init__(self):
        self.resource = None
        self.list = []
        self.NOTTED = 0

def IfParser(hdr):
    out = []
    i = 0
    while 1:
        m = IfHdr.search(hdr[i:])
        if not m: break

        i = i + m.end()
        tag = TagList()
        tag.resource = m.group('resource')
        if tag.resource:                # We need to delete < >
            tag.resource = tag.resource[1:-1]
        listitem = m.group('listitem')
        tag.NOTTED, tag.list = ListParser(listitem)
        out.append(tag)

    return out

def tokenFinder(token):
    # takes a string like '<opaquelocktoken:afsdfadfadf> and returns the token
    # part.
    if not token: return None           # An empty string was passed in
    if token[0] == '[': return None     # An Etag was passed in
    if token[0] == '<': token = token[1:-1]
    return token[token.find(':')+1:]

def ListParser(listitem):
    out = []
    NOTTED = 0
    i = 0
    while 1:
        m = ListItem.search(listitem[i:])
        if not m: break

        i = i + m.end()
        out.append(m.group('listitem'))
        if m.group('not'): NOTTED = 1

    return NOTTED, out



#!/usr/bin/env python

from __future__ import absolute_import
from setuptools import setup, find_packages
from io import open

VERSION = open('VERSION', 'r', encoding='utf-8').read()
VERSION = VERSION.replace('\n', '')

CHANGES = open('doc/Changes', 'r', encoding='utf-8').read()

DOC = """

Port from Andrew Leech PyWebDAV3 library to Support GNU Health.

It has been tested under Tryton server and Python3.

Some of the clients known to work:

- Mozilla Thunderbird (Lightning)
- Cadaver
- Konqueror
- Evolution

Changes
=======

%s
""" % CHANGES

setup(name='PyWebDAV3-GNUHealth',
      description='WebDAV library for Python3 - GNU Health port',
      author='GNU Solidario',
      author_email='health@gnusolidario.org',
      download_url='http://ftp.gnu.org/gnu/health/',
      maintainer='GNU Health team',
      maintainer_email='health@gnusolidario.org',
      url='http://health.gnu.org',
      platforms=['Unix'],
      license='GPL v3',
      version=VERSION,
      long_description=DOC,
      classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries',
        ],
      keywords=['webdav',
                'GNUHealth',
                'server',
                'dav',
                'standalone',
                'library',
                'gpl',
                'http',
                'rfc2518',
                'rfc 2518'
                ],
      packages=find_packages(),
      zip_safe=False,
      entry_points={
        'console_scripts': ['davserver = pywebdav.server.server:run']
        },
      )



# -*- coding: utf-8 -*-
##############################################################################
#
#    GNU Health: The Free Health and Hospital Information System
#    Copyright (C) 2008-2017 Luis Falcon <lfalcon@gnusolidario.org>
#    Copyright (C) 2011-2017 GNU Solidario <health@gnusolidario.org>
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import pytz
from dateutil.relativedelta import relativedelta
from trytond.model import ModelView, ModelSingleton, ModelSQL, fields
from datetime import datetime
from trytond.transaction import Transaction
from trytond import backend
from trytond.pool import Pool
from trytond.pyson import Eval, Not, Bool, PYSONEncoder, Equal, And, Or

__all__ = ['SurgerySequences', 'RCRI', 'Surgery', 'Operation',
    'SurgeryMainProcedure','SurgerySupply', 'PatientData', 'SurgeryTeam']



class SurgerySequences(ModelSingleton, ModelSQL, ModelView):
    "Inpatient Registration Sequences for GNU Health"
    __name__ = "gnuhealth.sequences"

    surgery_code_sequence = fields.Property(fields.Many2One(
        'ir.sequence', 'Surgery Sequence', required=True,
        domain=[('code', '=', 'gnuhealth.surgery')]))


class RCRI(ModelSQL, ModelView):
    'Revised Cardiac Risk Index'
    __name__ = 'gnuhealth.rcri'

    patient = fields.Many2One('gnuhealth.patient', 'Patient ID', required=True)
    rcri_date = fields.DateTime('Date', required=True)
    health_professional = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Professional',
        help="Health professional /"
        "Cardiologist who signed the assesment RCRI")

    rcri_high_risk_surgery = fields.Boolean(
        'High Risk surgery',
        help='Includes andy suprainguinal vascular, intraperitoneal,'
        ' or intrathoracic procedures')

    rcri_ischemic_history = fields.Boolean(
        'History of ischemic heart disease',
        help="history of MI or a positive exercise test, current \
        complaint of chest pain considered to be secondary to myocardial \
        ischemia, use of nitrate therapy, or ECG with pathological \
        Q waves; do not count prior coronary revascularization procedure \
        unless one of the other criteria for ischemic heart disease is \
        present")

    rcri_congestive_history = fields.Boolean(
        'History of congestive heart disease')

    rcri_diabetes_history = fields.Boolean(
        'Preoperative Diabetes',
        help="Diabetes Mellitus requiring treatment with Insulin")

    rcri_cerebrovascular_history = fields.Boolean(
        'History of Cerebrovascular disease')

    rcri_kidney_history = fields.Boolean(
        'Preoperative Kidney disease',
        help="Preoperative serum creatinine >2.0 mg/dL (177 mol/L)")

    rcri_total = fields.Integer(
        'Score',
        help='Points 0: Class I Very Low (0.4% complications)\n'
        'Points 1: Class II Low (0.9% complications)\n'
        'Points 2: Class III Moderate (6.6% complications)\n'
        'Points 3 or more : Class IV High (>11% complications)')

    rcri_class = fields.Selection([
        (None, ''),
        ('I', 'I'),
        ('II', 'II'),
        ('III', 'III'),
        ('IV', 'IV'),
        ], 'RCRI Class', sort=False)

    @fields.depends('rcri_high_risk_surgery', 'rcri_ischemic_history',
        'rcri_congestive_history', 'rcri_diabetes_history',
        'rcri_cerebrovascular_history', 'rcri_kidney_history')
    def on_change_with_rcri_total(self):

        total = 0
        if self.rcri_high_risk_surgery:
            total = total + 1
        if self.rcri_ischemic_history:
            total = total + 1
        if self.rcri_congestive_history:
            total = total + 1
        if self.rcri_diabetes_history:
            total = total + 1
        if self.rcri_kidney_history:
            total = total + 1
        if self.rcri_cerebrovascular_history:
            total = total + 1

        return total

    @fields.depends('rcri_high_risk_surgery', 'rcri_ischemic_history',
        'rcri_congestive_history', 'rcri_diabetes_history',
        'rcri_cerebrovascular_history', 'rcri_kidney_history')
    def on_change_with_rcri_class(self):
        rcri_class = ''

        total = 0
        if self.rcri_high_risk_surgery:
            total = total + 1
        if self.rcri_ischemic_history:
            total = total + 1
        if self.rcri_congestive_history:
            total = total + 1
        if self.rcri_diabetes_history:
            total = total + 1
        if self.rcri_kidney_history:
            total = total + 1
        if self.rcri_cerebrovascular_history:
            total = total + 1

        if total == 0:
            rcri_class = 'I'
        if total == 1:
            rcri_class = 'II'
        if total == 2:
            rcri_class = 'III'
        if (total > 2):
            rcri_class = 'IV'

        return rcri_class

    @staticmethod
    def default_rcri_date():
        return datetime.now()

    @staticmethod
    def default_rcri_total():
        return 0

    @staticmethod
    def default_rcri_class():
        return 'I'

    def get_rec_name(self, name):
        res = 'Points: ' + str(self.rcri_total) + ' (Class ' + \
            str(self.rcri_class) + ')'
        return res


class Surgery(ModelSQL, ModelView):
    'Surgery'
    __name__ = 'gnuhealth.surgery'


    def surgery_duration(self, name):

        if (self.surgery_end_date and self.surgery_date):
            return self.surgery_end_date - self.surgery_date
        else:
            return None

    def patient_age_at_surgery(self, name):
        if (self.patient.name.dob and self.surgery_date):
            rdelta = relativedelta (self.surgery_date.date(),
                self.patient.name.dob)
            years_months_days = str(rdelta.years) + 'y ' \
                + str(rdelta.months) + 'm ' \
                + str(rdelta.days) + 'd'
            return years_months_days
        else:
            return None

    patient = fields.Many2One('gnuhealth.patient', 'Patient', required=True)
    admission = fields.Many2One('gnuhealth.appointment', 'Admission')
    operating_room = fields.Many2One('gnuhealth.hospital.or', 'Operating Room')
    code = fields.Char('Code', readonly=True, help="Health Center code / sequence")

    procedures = fields.One2Many(
        'gnuhealth.operation', 'name', 'Procedures',
        help="List of the procedures in the surgery. Please enter the first "
        "one as the main procedure")

    supplies = fields.One2Many(
        'gnuhealth.surgery_supply', 'name', 'Supplies',
        help="List of the supplies required for the surgery")

    pathology = fields.Many2One(
        'gnuhealth.pathology', 'Condition',
        help="Base Condition / Reason")

    classification = fields.Selection([
        (None, ''),
        ('o', 'Optional'),
        ('r', 'Required'),
        ('u', 'Urgent'),
        ('e', 'Emergency'),
        ], 'Urgency', help="Urgency level for this surgery", sort=False)
    surgeon = fields.Many2One(
        'gnuhealth.healthprofessional', 'Surgeon',
        help="Surgeon who did the procedure")

    anesthetist = fields.Many2One(
        'gnuhealth.healthprofessional', 'Anesthetist',
        help="Anesthetist in charge")

    surgery_date = fields.DateTime(
        'Date', help="Start of the Surgery")

    surgery_end_date = fields.DateTime(
        'End',
        states={
            'required': Equal(Eval('state'), 'done'),
            },
        help="Automatically set when the surgery is done."
            "It is also the estimated end time when confirming the surgery.")

    surgery_length = fields.Function(
        fields.TimeDelta(
            'Duration',
            states={'invisible': And(Not(Equal(Eval('state'), 'done')),
                    Not(Equal(Eval('state'), 'signed')))},
            help="Length of the surgery"),
        'surgery_duration')

    state = fields.Selection([
        ('draft', 'Draft'),
        ('confirmed', 'Confirmed'),
        ('cancelled', 'Cancelled'),
        ('in_progress', 'In Progress'),
        ('done', 'Done'),
        ('signed', 'Signed'),
        ], 'State', readonly=True, sort=False)

    signed_by = fields.Many2One(
        'gnuhealth.healthprofessional', 'Signed by', readonly=True,
        states={
            'invisible': Not(Equal(Eval('state'), 'signed'))
            },

        help="Health Professional that signed this surgery document")

    # age is deprecated in GNU Health 2.0
    age = fields.Char(
        'Estimative Age',
        help="Use this field for historical purposes, \
        when no date of surgery is given")

    computed_age = fields.Function(
        fields.Char(
            'Age',
            help="Computed patient age at the moment of the surgery"),
        'patient_age_at_surgery')

    gender = fields.Function(fields.Selection([
        (None, ''),
        ('m', 'Male'),
        ('f', 'Female'),
        ('f-m','Female -> Male'),
        ('m-f','Male -> Female'),
        ], 'Gender'), 'get_patient_gender', searcher='search_patient_gender')


    description = fields.Char('Description')
    preop_mallampati = fields.Selection([
        (None, ''),
        ('Class 1', 'Class 1: Full visibility of tonsils, uvula and soft '
                    'palate'),
        ('Class 2', 'Class 2: Visibility of hard and soft palate, '
                    'upper portion of tonsils and uvula'),
        ('Class 3', 'Class 3: Soft and hard palate and base of the uvula are '
                    'visible'),
        ('Class 4', 'Class 4: Only Hard Palate visible'),
        ], 'Mallampati Score', sort=False)
    preop_bleeding_risk = fields.Boolean(
        'Risk of Massive bleeding',
        help="Patient has a risk of losing more than 500 "
        "ml in adults of over 7ml/kg in infants. If so, make sure that "
        "intravenous access and fluids are available")

    preop_oximeter = fields.Boolean(
        'Pulse Oximeter in place',
        help="Pulse oximeter is in place "
        "and functioning")

    preop_site_marking = fields.Boolean(
        'Surgical Site Marking',
        help="The surgeon has marked the surgical incision")

    preop_antibiotics = fields.Boolean(
        'Antibiotic Prophylaxis',
        help="Prophylactic antibiotic treatment within the last 60 minutes")

    preop_sterility = fields.Boolean(
        'Sterility confirmed',
        help="Nursing team has confirmed sterility of the devices and room")

    preop_asa = fields.Selection([
        (None, ''),
        ('ps1', 'PS 1 : Normal healthy patient'),
        ('ps2', 'PS 2 : Patients with mild systemic disease'),
        ('ps3', 'PS 3 : Patients with severe systemic disease'),
        ('ps4', 'PS 4 : Patients with severe systemic disease that is'
            ' a constant threat to life '),
        ('ps5', 'PS 5 : Moribund patients who are not expected to'
            ' survive without the operation'),
        ('ps6', 'PS 6 : A declared brain-dead patient who organs are'
            ' being removed for donor purposes'),
        ], 'ASA PS',
        help="ASA pre-operative Physical Status", sort=False)

    preop_rcri = fields.Many2One(
        'gnuhealth.rcri', 'RCRI',
        help='Patient Revised Cardiac Risk Index\n'
        'Points 0: Class I Very Low (0.4% complications)\n'
        'Points 1: Class II Low (0.9% complications)\n'
        'Points 2: Class III Moderate (6.6% complications)\n'
        'Points 3 or more : Class IV High (>11% complications)')

    surgical_wound = fields.Selection([
        (None, ''),
        ('I', 'Clean . Class I'),
        ('II', 'Clean-Contaminated . Class II'),
        ('III', 'Contaminated . Class III'),
        ('IV', 'Dirty-Infected . Class IV'),
        ], 'Surgical wound', sort=False)


    extra_info = fields.Text('Extra Info')

    anesthesia_report = fields.Text('Anesthesia Report')

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    report_surgery_date = fields.Function(fields.Date('Surgery Date'), 
        'get_report_surgery_date')
    report_surgery_time = fields.Function(fields.Time('Surgery Time'), 
        'get_report_surgery_time')

    surgery_team = fields.One2Many(
        'gnuhealth.surgery_team', 'name', 'Team Members',
        help="Professionals Involved in the surgery")

    postoperative_dx = fields.Many2One(
        'gnuhealth.pathology', 'Post-op dx',
        states={'invisible': And(Not(Equal(Eval('state'), 'done')),
                    Not(Equal(Eval('state'), 'signed')))},
        help="Post-operative diagnosis")

    @staticmethod
    def default_institution():
        HealthInst = Pool().get('gnuhealth.institution')
        institution = HealthInst.get_institution()
        return institution

    @staticmethod
    def default_surgery_date():
        return datetime.now()

    @staticmethod
    def default_surgeon():
        pool = Pool()
        HealthProf= pool.get('gnuhealth.healthprofessional')
        surgeon = HealthProf.get_health_professional()
        return surgeon

    @staticmethod
    def default_state():
        return 'draft'

    def get_patient_gender(self, name):
        return self.patient.gender

    @classmethod
    def search_patient_gender(cls, name, clause):
        res = []
        value = clause[2]
        res.append(('patient.name.gender', clause[1], value))
        return res


    # Show the gender and age upon entering the patient 
    # These two are function fields (don't exist at DB level)
    @fields.depends('patient')
    def on_change_patient(self):
        gender=None
        age=''
        self.gender = self.patient.gender
        self.computed_age = self.patient.age


    @classmethod
    def create(cls, vlist):
        Sequence = Pool().get('ir.sequence')
        Config = Pool().get('gnuhealth.sequences')

        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('code'):
                config = Config(1)
                values['code'] = Sequence.get_id(
                    config.surgery_code_sequence.id)
        return super(Surgery, cls).create(vlist)


    @classmethod
    def __setup__(cls):
        super(Surgery, cls).__setup__()
        cls._error_messages.update({
            'end_date_before_start': 'End time "%(end_date)s" BEFORE '
                'surgery date "%(surgery_date)s"',
            'or_is_not_available': 'Operating Room is not available'})

        cls._buttons.update({
            'confirmed': {
                'invisible': And(Not(Equal(Eval('state'), 'draft')),
                    Not(Equal(Eval('state'), 'cancelled'))),
                },
            'cancel': {
                'invisible': Not(Equal(Eval('state'), 'confirmed')),
                },
            'start': {
                'invisible': Not(Equal(Eval('state'), 'confirmed')),
                },
            'done': {
                'invisible': Not(Equal(Eval('state'), 'in_progress')),
                },
            'signsurgery': {
                'invisible': Not(Equal(Eval('state'), 'done')),
                },

            })
        
    @classmethod
    def validate(cls, surgeries):
        super(Surgery, cls).validate(surgeries)
        for surgery in surgeries:
            surgery.validate_surgery_period()

    def validate_surgery_period(self):
        Lang = Pool().get('ir.lang')

        language, = Lang.search([
            ('code', '=', Transaction().language),
            ])
        if (self.surgery_end_date and self.surgery_date):
            if (self.surgery_end_date < self.surgery_date):
                self.raise_user_error('end_date_before_start', {
                        'surgery_date': Lang.strftime(self.surgery_date,
                            language.code,
                            language.date),
                        'end_date': Lang.strftime(self.surgery_end_date,
                            language.code,
                            language.date),
                        })


    @classmethod
    def write(cls, surgeries, vals):
        # Don't allow to write the record if the surgery has been signed
        if surgeries[0].state == 'signed':
            cls.raise_user_error(
                "This surgery is at state Done and has been signed\n"
                "You can no longer modify it.")
        return super(Surgery, cls).write(surgeries, vals)

    ## Method to check for availability and make the Operating Room reservation
     # for the associated surgery
     
    @classmethod
    @ModelView.button
    def confirmed(cls, surgeries):
        surgery_id = surgeries[0]
        Operating_room = Pool().get('gnuhealth.hospital.or')
        cursor = Transaction().connection.cursor()

        # Operating Room and end surgery time check
        if (not surgery_id.operating_room or not surgery_id.surgery_end_date):
            cls.raise_user_error("Operating Room and estimated end time  "
            "are needed in order to confirm the surgery")

        or_id = surgery_id.operating_room.id
        cursor.execute("SELECT COUNT(*) \
            FROM gnuhealth_surgery \
            WHERE (surgery_date::timestamp,surgery_end_date::timestamp) \
                OVERLAPS (timestamp %s, timestamp %s) \
              AND (state = %s or state = %s) \
              AND operating_room = CAST(%s AS INTEGER) ",
            (surgery_id.surgery_date,
            surgery_id.surgery_end_date,
            'confirmed', 'in_progress', str(or_id)))
        res = cursor.fetchone()
        if (surgery_id.surgery_end_date <
            surgery_id.surgery_date):
            cls.raise_user_error("The Surgery end date must later than the \
                Start")
        if res[0] > 0:
            cls.raise_user_error('or_is_not_available')
        else:
            cls.write(surgeries, {'state': 'confirmed'})
 
    # Cancel the surgery and set it to draft state
    # Free the related Operating Room
    
    @classmethod
    @ModelView.button
    def cancel(cls, surgeries):
        surgery_id = surgeries[0]
        Operating_room = Pool().get('gnuhealth.hospital.or')
        
        cls.write(surgeries, {'state': 'cancelled'})

    # Start the surgery
    
    @classmethod
    @ModelView.button
    def start(cls, surgeries):
        surgery_id = surgeries[0]
        Operating_room = Pool().get('gnuhealth.hospital.or')

        cls.write(surgeries, 
            {'state': 'in_progress',
             'surgery_date': datetime.now(),
             'surgery_end_date': datetime.now()})
        Operating_room.write([surgery_id.operating_room], {'state': 'occupied'})


    # Finnish the surgery
    # Free the related Operating Room
    
    @classmethod
    @ModelView.button
    def done(cls, surgeries):
        surgery_id = surgeries[0]
        Operating_room = Pool().get('gnuhealth.hospital.or')
        
        cls.write(surgeries, {'state': 'done',
                              'surgery_end_date': datetime.now()})
                              
        Operating_room.write([surgery_id.operating_room], {'state': 'free'})


    # Sign the surgery document, and the surgical act.

    @classmethod
    @ModelView.button
    def signsurgery(cls, surgeries):
        surgery_id = surgeries[0]

        # Sign, change the state of the Surgery to "Signed"
        # and write the name of the signing health professional

        signing_hp = Pool().get('gnuhealth.healthprofessional').get_health_professional()
        if not signing_hp:
            cls.raise_user_error(
                "No health professional associated to this user !")

        cls.write(surgeries, {
            'state': 'signed',
            'signed_by': signing_hp})

    def get_report_surgery_date(self, name):
        Company = Pool().get('company.company')

        timezone = None
        company_id = Transaction().context.get('company')
        if company_id:
            company = Company(company_id)
            if company.timezone:
                timezone = pytz.timezone(company.timezone)

        dt = self.surgery_date
        return datetime.astimezone(dt.replace(tzinfo=pytz.utc), timezone).date()

    def get_report_surgery_time(self, name):
        Company = Pool().get('company.company')

        timezone = None
        company_id = Transaction().context.get('company')
        if company_id:
            company = Company(company_id)
            if company.timezone:
                timezone = pytz.timezone(company.timezone)

        dt = self.surgery_date
        return datetime.astimezone(dt.replace(tzinfo=pytz.utc), timezone).time()




class Operation(ModelSQL, ModelView):
    'Operation - Surgical Procedures'
    __name__ = 'gnuhealth.operation'

    name = fields.Many2One('gnuhealth.surgery', 'Surgery')
    procedure = fields.Many2One(
        'gnuhealth.procedure', 'Code', required=True, select=True,
        help="Procedure Code, for example ICD-10-PCS or ICPM")
    notes = fields.Text('Notes')

    def get_rec_name(self, name):
        return self.procedure.rec_name


class SurgeryMainProcedure(ModelSQL, ModelView):
    __name__ = 'gnuhealth.surgery'

    main_procedure = fields.Many2One('gnuhealth.operation','Main Proc',
        domain=[('name', '=', Eval('active_id'))],
        states={'readonly': Or(~Eval('procedures'), Eval('id', 0) < 0)},
        depends=['procedures'])

class SurgerySupply(ModelSQL, ModelView):
    'Supplies related to the surgery'
    __name__ = 'gnuhealth.surgery_supply'

    name = fields.Many2One('gnuhealth.surgery', 'Surgery')
    qty = fields.Numeric('Qty',required=True,
        help="Initial required quantity")
    supply = fields.Many2One(
        'product.product', 'Supply', required=True,
        domain=[('is_medical_supply', '=', True)],
        help="Supply to be used in this surgery")
   
    notes = fields.Char('Notes')
    qty_used = fields.Numeric('Used', required=True,
        help="Actual amount used")
    
class SurgeryTeam(ModelSQL, ModelView):
    'Team Involved in the surgery'
    __name__ = 'gnuhealth.surgery_team'

    name = fields.Many2One('gnuhealth.surgery', 'Surgery')
    team_member = fields.Many2One(
        'gnuhealth.healthprofessional', 'Member', required=True, select=True,
        help="Health professional that participated on this surgery")

    role = fields.Many2One(
        'gnuhealth.hp_specialty', 'Role',
        domain=[('name', '=', Eval('team_member'))],
        depends=['team_member'])
    
    notes = fields.Char('Notes')

class PatientData(ModelSQL, ModelView):
    __name__ = 'gnuhealth.patient'

    surgery = fields.One2Many(
        'gnuhealth.surgery', 'patient', 'Surgeries', readonly=True)



# -*- coding: utf-8 -*-
##############################################################################
#
#    GNU Health: The Free Health and Hospital Information System
#    Copyright (C) 2008-2017 Luis Falcon <lfalcon@gnu.org>
#    Copyright (C) 2013  Sebastián Marro <smarro@thymbra.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from datetime import datetime
from trytond.model import Workflow, ModelView, ModelSingleton, ModelSQL, fields
from trytond.pyson import Eval
from trytond.pool import Pool


__all__ = [
    'GnuHealthSequences', 'ImagingTestType', 'ImagingTest',
    'ImagingTestRequest', 'ImagingTestResult']


class GnuHealthSequences(ModelSingleton, ModelSQL, ModelView):
    "Standard Sequences for GNU Health"
    __name__ = "gnuhealth.sequences"

    imaging_request_sequence = fields.Property(
        fields.Many2One(
            'ir.sequence',
            'Imaging Request Sequence',
            domain=[('code', '=', 'gnuhealth.imaging.test.request')],
            required=True))

    imaging_sequence = fields.Property(
        fields.Many2One(
            'ir.sequence',
            'Imaging Sequence',
            domain=[('code', '=', 'gnuhealth.imaging.test.result')],
            required=True))


class ImagingTestType(ModelSQL, ModelView):
    'Imaging Test Type'
    __name__ = 'gnuhealth.imaging.test.type'

    code = fields.Char('Code', required=True)
    name = fields.Char('Name', required=True)


class ImagingTest(ModelSQL, ModelView):
    'Imaging Test'
    __name__ = 'gnuhealth.imaging.test'

    code = fields.Char('Code', required=True)
    name = fields.Char('Name', required=True)
    test_type = fields.Many2One(
        'gnuhealth.imaging.test.type', 'Type',
        required=True)
    product = fields.Many2One('product.product', 'Service', required=True)

    active = fields.Boolean('Active', select=True)

    @staticmethod
    def default_active():
        return True

class ImagingTestRequest(Workflow, ModelSQL, ModelView):
    'Imaging Test Request'
    __name__ = 'gnuhealth.imaging.test.request'

    patient = fields.Many2One('gnuhealth.patient', 'Patient', required=True)
    date = fields.DateTime('Date', required=True)
    requested_test = fields.Many2One(
        'gnuhealth.imaging.test', 'Test',
        required=True)
    doctor = fields.Many2One('gnuhealth.healthprofessional', 'Doctor', required=True)
    state = fields.Selection([
        ('draft', 'Draft'),
        ('requested', 'Requested'),
        ('done', 'Done'),
        ], 'State', readonly=True)
    comment = fields.Text('Comment')
    request = fields.Char('Request', readonly=True)
    urgent = fields.Boolean('Urgent')

    @classmethod
    def __setup__(cls):
        super(ImagingTestRequest, cls).__setup__()
        cls._transitions |= set((
            ('draft', 'requested'),
            ('requested', 'done')
        ))
        cls._buttons.update({
            'requested': {
                'invisible': ~Eval('state').in_(['draft']),
                },
            'generate_results': {
                'invisible': ~Eval('state').in_(['requested'])
                }
            })
        cls._order.insert(0, ('date', 'DESC'))
        cls._order.insert(1, ('request', 'DESC'))

    @staticmethod
    def default_date():
        return datetime.now()

    @staticmethod
    def default_state():
        return 'draft'

    @staticmethod
    def default_doctor():
        pool = Pool()
        HealthProf= pool.get('gnuhealth.healthprofessional')
        hp = HealthProf.get_health_professional()
        return hp

    @classmethod
    def create(cls, vlist):
        Sequence = Pool().get('ir.sequence')
        Config = Pool().get('gnuhealth.sequences')

        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('request'):
                config = Config(1)
                values['request'] = Sequence.get_id(
                    config.imaging_request_sequence.id)

        return super(ImagingTestRequest, cls).create(vlist)

    @classmethod
    def copy(cls, tests, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default['request'] = None
        default['date'] = cls.default_date()
        return super(ImagingTestRequest, cls).copy(tests, default=default)

    @classmethod
    @ModelView.button
    @Workflow.transition('requested')
    def requested(cls, requests):
        pass

    @classmethod
    @ModelView.button_action('health_imaging.wizard_generate_result')
    def generate_results(cls, requests):
        pass

    @classmethod
    @Workflow.transition('done')
    def done(cls, requests):
        pass


class ImagingTestResult(ModelSQL, ModelView):
    'Imaging Test Result'
    __name__ = 'gnuhealth.imaging.test.result'

    patient = fields.Many2One('gnuhealth.patient', 'Patient', readonly=True)
    number = fields.Char('Number', readonly=True)
    date = fields.DateTime('Date', required=True)
    request_date = fields.DateTime('Requested Date', readonly=True)
    requested_test = fields.Many2One(
        'gnuhealth.imaging.test', 'Test',
        required=True)
    request = fields.Many2One(
        'gnuhealth.imaging.test.request', 'Request',
        readonly=True)
    doctor = fields.Many2One('gnuhealth.healthprofessional', 'Doctor', required=True)
    comment = fields.Text('Comment')
    images = fields.One2Many('ir.attachment', 'resource', 'Images')

    @classmethod
    def create(cls, vlist):
        Sequence = Pool().get('ir.sequence')
        Config = Pool().get('gnuhealth.sequences')

        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('name'):
                config = Config(1)
                values['number'] = Sequence.get_id(
                    config.imaging_sequence.id)

        return super(ImagingTestResult, cls).create(vlist)



"""

basic interface class

use this for subclassing when writing your own interface
class.

"""


from xml.dom import minidom
from .locks import LockManager
from .errors import *

import time

class dav_interface:
    """ interface class for implementing DAV servers """

    ### defined properties (modify this but let the DAV stuff there!)
    ### the format is namespace: [list of properties]

    PROPS={"DAV:" : ('creationdate',
                     'displayname',
                     'getcontentlanguage',
                     'getcontentlength',
                     'getcontenttype',
                     'getetag',
                     'getlastmodified',
                     'lockdiscovery',
                     'resourcetype',
                     'source',
                     'supportedlock'),
           "NS2" : ("p1","p2")
           }

    # here we define which methods handle which namespace
    # the first item is the namespace URI and the second one
    # the method prefix
    # e.g. for DAV:getcontenttype we call dav_getcontenttype()
    M_NS={"DAV:" : "_get_dav",
          "NS2"  : "ns2" }

    def get_propnames(self,uri):
        """ return the property names allowed for the given URI

        In this method we simply return the above defined properties
        assuming that they are valid for any resource.
        You can override this in order to return a different set
        of property names for each resource.

        """
        return self.PROPS

    def get_prop2(self,uri,ns,pname):
        """ return the value of a property

        """
        if ns.lower() == "dav:":
            return self.get_dav(uri,pname)

        raise DAV_NotFound

    def get_prop(self,uri,ns,propname):
        """ return the value of a given property

        uri        -- uri of the object to get the property of
        ns        -- namespace of the property
        pname        -- name of the property
        """
        if ns in self.M_NS:
            prefix=self.M_NS[ns]
        else:
            raise DAV_NotFound
        mname=prefix+"_"+propname.replace('-', '_')
        try:
            m=getattr(self,mname)
            r=m(uri)
            return r
        except AttributeError:
            raise DAV_NotFound

    ###
    ### DATA methods (for GET and PUT)
    ###

    def get_data(self, uri, range=None):
        """ return the content of an object

        return data or raise an exception

        """
        raise DAV_NotFound

    def put(self, uri, data, content_type=None):
        """ write an object to the repository

        return the location uri or raise an exception
        """

        raise DAV_Forbidden

    ###
    ### LOCKing information
    ###
    def _get_dav_supportedlock(self, uri):

        txt = ('<main xmlns:D="http://dummy/d" xmlns:ns1="http://webdav.de/ns1"><D:lockentry>\n'
                '<D:lockscope><D:exclusive></D:exclusive></D:lockscope>\n'
                '<D:locktype><D:write></D:write></D:locktype>\n'
                '</D:lockentry></main>\n')
        xml = minidom.parseString(txt)
        return xml.firstChild.firstChild

    def _get_dav_lockdiscovery(self, uri):
        lcm = LockManager()
        if lcm._l_isLocked(uri):
            lock = lcm._l_getLockForUri(uri)
            txt = lock.asXML(discover=True, namespace='D')

            txtwithns = '<main xmlns:D="http://dummy/D" xmlns:n="http://webdav.de/N">%s</main>'
            xml = minidom.parseString(txtwithns % txt)
            return xml.firstChild.firstChild

        return ''

    ###
    ### Methods for DAV properties
    ###

    def _get_dav_creationdate(self,uri):
        """ return the creationdate of a resource """
        d=self.get_creationdate(uri)
        # format it
        return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(d))

    def _get_dav_getlastmodified(self,uri):
        """ return the last modified date of a resource """
        d=self.get_lastmodified(uri)
        # format it
        return time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime(d))


    ###
    ### OVERRIDE THESE!
    ###

    def get_creationdate(self,uri):
        """ return the creationdate of the resource """
        return time.time()

    def get_lastmodified(self,uri):
        """ return the last modification date of the resource """
        return time.time()


    ###
    ### COPY MOVE DELETE
    ###

    ### methods for deleting a resource

    def rmcol(self,uri):
        """ delete a collection 

        This should not delete any children! This is automatically done
        before by the DELETE class in DAV/delete.py

        return a success code or raise an exception

        """
        raise DAV_NotFound

    def rm(self,uri):
        """ delete a single resource 

        return a success code or raise an exception

        """
        raise DAV_NotFound

    """

    COPY/MOVE HANDLER

    These handler are called when a COPY or MOVE method is invoked by
    a client. In the default implementation it works as follows:

    - the davserver receives a COPY/MOVE method
    - the davcopy or davmove module will be loaded and the corresponding
      class will be initialized
    - this class parses the query and decides which method of the interface class
      to call:

      copyone for a single resource to copy
      copytree for a tree to copy (collection)
      (the same goes for move of course).

    - the interface class has now two options:
        1. to handle the action directly (e.g. cp or mv on filesystems)
        2. to let it handle via the copy/move methods in davcmd.

    ad 1) The first approach can be used when we know that no error can 
          happen inside a tree or when the action can exactly tell which
          element made which error. We have to collect these and return
          it in a dict of the form {uri: error_code, ...}

    ad 2) The copytree/movetree/... methods of davcmd.py will do the recursion
          themselves and call for each resource the copy/move method of the
          interface class. Thus method will then only act on a single resource.
          (Thus a copycol on a normal unix filesystem actually only needs to do
          an mkdir as the content will be copied by the davcmd.py function.
          The davcmd.py method will also automatically collect all errors and
          return the dictionary described above.
          When you use 2) you also have to implement the copy() and copycol()
          methods in your interface class. See the example for details.

    To decide which approach is the best you have to decide if your application
    is able to generate errors inside a tree. E.g. a function which completely
    fails on a tree if one of the tree's childs fail is not what we need. Then
    2) would be your way of doing it.
    Actually usually 2) is the better solution and should only be replaced by
    1) if you really need it.

    The remaining question is if we should do the same for the DELETE method.

    """

    ### MOVE handlers

    def moveone(self,src,dst,overwrite):
        """ move one resource with Depth=0 """
        return moveone(self,src,dst,overwrite)

    def movetree(self,src,dst,overwrite):
        """ move a collection with Depth=infinity """
        return movetree(self,src,dst,overwrite)

    ### COPY handlers

    def copyone(self,src,dst,overwrite):
        """ copy one resource with Depth=0 """
        return copyone(self,src,dst,overwrite)

    def copytree(self,src,dst,overwrite):
        """ copy a collection with Depth=infinity """
        return copytree(self,src,dst,overwrite)


    ### low level copy methods (you only need these for method 2)
    def copy(self,src,dst):
        """ copy a resource with depth==0 

        You don't need to bother about overwrite or not.
        This has been done already.

        return a success code or raise an exception if something fails
        """
        return 201


    def copycol(self,src,dst):
        """ copy a resource with depth==infinity 

        You don't need to bother about overwrite or not.
        This has been done already.

        return a success code or raise an exception if something fails
        """
        return 201

    ### some utility functions you need to implement

    def exists(self,uri):
        """ return 1 or None depending on if a resource exists """
        return None # no

    def is_collection(self,uri):
        """ return 1 or None depending on if a resource is a collection """
        return None # no




class immunizationRoute:
    contents=[{'code': 'IM',
                'display': 'Intramuscular'},
            {'code': 'PO',
                'display': 'Oral'},
            {'code': 'NAS',
                'display': 'Intranasal'},
            {'code': 'SC',
                'display': 'Subcutaneous'},
            {'code': 'ID',
                'display': 'Intradermal'}]

__all__=['immunizationRoute']
